@@ -52,15 +52,15 @@
 #define LOG_SUBSYSTEM "font_unifont"
 
 /*
- * Glyph data is linked to the binary externally as binary data. The data layout
- * is a size-byte followed by 32 data bytes. The data bytes are padded with 0 if
- * the size is smaller than 32.
- * Sizes bigger than 32 are not used.
+ * We regroup all glyphs into blocks, of contiguous codepoints, and same width.
+ * This allows to better pack the data, and handle some codepoints that are
+ * not in the 0xffff range
  */
-
-struct unifont_data {
-	uint8_t len;
-	uint8_t data[32];
+struct unifont_glyph_block {
+	uint32_t codepoint; // First codepoint of the block
+	uint32_t offset;    // offset of the data
+	uint16_t len;	    // number of glyph in this block
+	uint8_t width;	    // glyph width (1 or 2 for double-width glyph)
 } __attribute__((__packed__));
 
 /*
@@ -99,18 +99,68 @@ static void free_glyph(void *data)
 	free(g);
 }
 
-static void unfold(uint8_t *dst, uint8_t val)
+static uint8_t apply_attr(uint8_t c, const struct kmscon_font_attr *attr, bool last_line)
 {
-	*dst = 0xff * !!val;
+	if (attr->bold)
+		c |= c >> 1;
+	if (attr->underline && last_line)
+		c = 0xff;
+	return c;
 }
 
-static int find_glyph(uint32_t ch, const struct kmscon_glyph **out)
+static uint8_t unfold(uint8_t val)
+{
+	return 0xff * !!val;
+}
+
+static bool is_in_block(const struct unifont_glyph_block *idx, uint32_t ch)
+{
+	return (ch >= idx->codepoint && ch < idx->codepoint + idx->len);
+}
+
+static int lookup_block(const struct unifont_glyph_block *blocks, uint32_t len, uint32_t ch)
+{
+	int look = 1 + ((ch * len) / 0xffff); /* opportunist first look*/
+	int min = 0;
+	int max = len - 1;
+
+	if (look > max)
+		look = max;
+
+	while (min != max) {
+		log_debug("lookup %d codep %d, look %d min %d max %d", ch, blocks[look].codepoint,
+			  look, min, max);
+
+		if (is_in_block(&blocks[look], ch))
+			return look;
+		if (ch < blocks[look].codepoint) {
+			max = look;
+			look -= look - min > 2 ? (look - min) / 2 : 1;
+
+		} else {
+			min = look;
+			look += max - look > 2 ? (max - look) / 2 : 1;
+		}
+	}
+	if (is_in_block(&blocks[look], ch))
+		return look;
+	return -1;
+}
+
+static int find_glyph(uint64_t id, const struct kmscon_glyph **out,
+		      const struct kmscon_font_attr *attr)
 {
 	struct kmscon_glyph *g;
+	uint32_t ch = id & TSM_UCS4_MAX;
+	const void *start = _binary_font_unifont_data_start;
+	const uint8_t *end = (uint8_t *)_binary_font_unifont_data_end;
+	const uint8_t *data;
+	uint8_t c;
+	uint32_t len;
+	const struct unifont_glyph_block *blocks;
+	unsigned int i, k;
 	int ret;
 	bool res;
-	const struct unifont_data *start, *end, *d;
-	unsigned int i, w;
 
 	pthread_mutex_lock(&cache_mutex);
 
@@ -121,49 +171,48 @@ static int find_glyph(uint32_t ch, const struct kmscon_glyph **out)
 			goto out_unlock;
 		}
 	} else {
-		res = shl_hashtable_find(cache, (void **)out, ch);
+		res = shl_hashtable_find(cache, (void **)out, id);
 		if (res) {
 			ret = 0;
 			goto out_unlock;
 		}
 	}
+	/* First the length of the block index */
+	len = *((uint32_t *)start);
+	/* Then the block index */
+	blocks = (struct unifont_glyph_block *)(start + 4);
+	/* Then the glyph data */
+	data = (uint8_t *)start + 4 + len * sizeof(struct unifont_glyph_block);
 
-	if (ch > 0xffff) {
-		ret = -ERANGE;
-		goto out_unlock;
+	int idx = lookup_block(blocks, len, ch);
+	if (idx < 0) {
+		log_debug("codepoint %08x not found, using replacement glyph", ch);
+		ch = 0xfffd;
+		idx = lookup_block(blocks, len, ch);
+		if (idx < 0) {
+			log_warning("Replacement glyph not found");
+			ret = -1;
+			goto out_unlock;
+		}
 	}
 
-	start = (const struct unifont_data *)_binary_font_unifont_data_start;
-	end = (const struct unifont_data *)_binary_font_unifont_data_end;
-	d = &start[ch];
-
-	if (d >= end) {
+	data += blocks[idx].offset + (ch - blocks[idx].codepoint) * blocks[idx].width * 16;
+	if (data + 16 * blocks[idx].width > end) {
+		log_warning("glyph out of range %p %p", data, end);
 		ret = -ERANGE;
 		goto out_unlock;
 	}
 
-	switch (d->len) {
-	case 16:
-		w = 1;
-		break;
-	case 32:
-		w = 2;
-		break;
-	default:
-		ret = -EFAULT;
-		goto out_unlock;
-	}
-
 	g = malloc(sizeof(*g));
 	if (!g) {
 		ret = -ENOMEM;
 		goto out_unlock;
 	}
 	memset(g, 0, sizeof(*g));
-	g->width = w;
-	g->buf.width = w * 8;
+	g->width = blocks[idx].width;
+	g->buf.width = g->width * 8;
 	g->buf.height = 16;
-	g->buf.stride = w * 8;
+	g->buf.stride = g->width * 8;
 	g->buf.format = UTERM_FORMAT_GREY;
 
 	g->buf.data = malloc(g->buf.stride * g->buf.height);
@@ -172,18 +221,13 @@ static int find_glyph(uint32_t ch, const struct kmscon_glyph **out)
 		goto err_free;
 	}
 
-	for (i = 0; i < d->len; ++i) {
-		unfold(&g->buf.data[i * 8 + 0], d->data[i] & 0x80);
-		unfold(&g->buf.data[i * 8 + 1], d->data[i] & 0x40);
-		unfold(&g->buf.data[i * 8 + 2], d->data[i] & 0x20);
-		unfold(&g->buf.data[i * 8 + 3], d->data[i] & 0x10);
-		unfold(&g->buf.data[i * 8 + 4], d->data[i] & 0x08);
-		unfold(&g->buf.data[i * 8 + 5], d->data[i] & 0x04);
-		unfold(&g->buf.data[i * 8 + 6], d->data[i] & 0x02);
-		unfold(&g->buf.data[i * 8 + 7], d->data[i] & 0x01);
+	for (i = 0; i < g->width * 16; ++i) {
+		c = apply_attr(data[i], attr, i >= g->width * 15);
+		for (k = 0; k < 8; k++)
+			g->buf.data[i * 8 + k] = unfold(c & (1 << (7 - k)));
 	}
 
-	ret = shl_hashtable_insert(cache, ch, g);
+	ret = shl_hashtable_insert(cache, id, g);
 	if (ret) {
 		log_error("cannot insert glyph into glyph-cache: %d", ret);
 		goto err_data;
@@ -215,12 +259,11 @@ static int kmscon_font_unifont_init(struct kmscon_font *out, const struct kmscon
 
 	memset(&out->attr, 0, sizeof(out->attr));
 	memcpy(out->attr.name, name, sizeof(name));
-	out->attr.bold = false;
+	out->attr.bold = attr->bold;
 	out->attr.italic = false;
 	out->attr.width = 8;
 	out->attr.height = 16;
 	kmscon_font_attr_normalize(&out->attr);
-	out->baseline = 4;
 
 	cache_ref();
 	return 0;
@@ -238,19 +281,19 @@ static int kmscon_font_unifont_render(struct kmscon_font *font, uint64_t id, con
 	if (len > 1)
 		return -ERANGE;
 
-	return find_glyph(id & TSM_UCS4_MAX, out);
+	return find_glyph(id, out, &font->attr);
 }
 
 static int kmscon_font_unifont_render_inval(struct kmscon_font *font,
 					    const struct kmscon_glyph **out)
 {
-	return find_glyph(0xfffd, out);
+	return find_glyph(0xfffd, out, &font->attr);
 }
 
 static int kmscon_font_unifont_render_empty(struct kmscon_font *font,
 					    const struct kmscon_glyph **out)
 {
-	return find_glyph(' ', out);
+	return find_glyph(' ', out, &font->attr);
 }
 
 struct kmscon_font_ops kmscon_font_unifont_ops = {
